from . import collectible
from . import config
from . import context
from . import criterion
from . import data_builder
from . import dataset
from . import exceptions
from . import experiment
from . import loader
from . import metric
from . import model
from . import optimizer
from . import registry
from . import writer
from . import transform
from . import search
from . import parameter
from . import nn



from lighter.decorator import dataset, context


class BaseDataBuilder(object):
    """
    Base class for building the data loaders with injected 'dataset' reference and application 'context'.
    """
    @dataset
    @context
    def __init__(self):
        pass

    def loader(self, *args, **kwargs):
        """
        Returns the data loader instances.
        :param args:
        :param kwargs:
        :return: data loader(s) for an experiment.
        """
        raise NotImplementedError('Cannot instantiate builder base class.')



import functools
import logging
import inspect
from typing import Tuple, List, Callable
from enum import Enum
from box import Box
from lighter.config import Config
from lighter.context import Context
from lighter.exceptions import DependencyInjectionError
from lighter.loader import Loader
from lighter.misc import DotDict
from lighter.registry import Registry
from lighter.search import ParameterSearch
from lighter.parameter import Parameter

DEFAULT_PROPERTIES = ['transform', 'dataset', 'data_builder', 'model', 'optimizer',
                      'collectible', 'criterion', 'metric', 'writer']


def _handle_injections(args, injectables: List[str], ignore_none_values):
    registry = Registry.get_instance()
    instance = args[0]

    for inject in injectables:
        parent, name = DotDict.resolve(registry.instances, inject)
        value = getattr(parent, name, None)
        if not ignore_none_values and value is None:
            raise DependencyInjectionError('Trying to inject dependency of unresolved '
                                           'key: {} into instance: {} with value: {}'
                                           .format(inject, instance, value))
        setattr(instance, name, value)


def _handle_config(args, path, source):
    context = Context.get_instance()
    config = Config.get_instance()
    instance = args[0]
    # check if context is allowed to change and perform config updates
    if context.allow_context_changes and path is not None:
        imported_config, _ = Config.load(path=path)
        if source is not None:
            parent, name = DotDict.resolve(config, source)
            setattr(parent, name, imported_config)
        else:
            for k, v in imported_config.items():
                setattr(config, k, v)
    setattr(instance, 'config', config)


def _handle_registration(source):
    context = Context.get_instance()
    # prevent type registration is context is not allowed to change
    if not context.allow_context_changes:
        return
    config = Config.get_instance()
    parent, name = DotDict.resolve(config, source)
    types = getattr(parent, name)
    types = {k: Loader.import_path(v[len("type::"):]) for k, v in types.strategy.items()}
    context.instantiate_types(types)


def _wrapper_delegate(func, injectables, ignore_none_values):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        _handle_injections(args, injectables, ignore_none_values)
        return func(*args, **kwargs)
    return wrapper


def _search_and_load_type(instance):
    registry = Registry.get_instance()
    reg_instance, key = DotDict.resolve(registry.types, instance)
    class_ = getattr(reg_instance, key, None)
    if class_ is None or not inspect.isclass(class_):
        raise DependencyInjectionError()
    new_ = class_()
    registry.register_instance(key, new_)
    return new_


# ----------- PUBLIC DECORATORS -----------


def device(func=None, name: str = None, source: str = 'device.default', property: str = 'device'):
    """
    Injects a device info into the current object instance.
    :param func: original function instance
    :param name: device id if pre-specified
    :param source: source path for the config
    :param property: device name for injection
    :return:
    """
    if not func:
        # workaround to enable empty decorator
        return functools.partial(device, name=name, source=source, property=property)

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        context = Context.get_instance()
        config = Config.get_instance()
        obj_instance = args[0]

        # update device according to id and check if context is allowed to be updated
        if context.allow_context_changes and name is not None:
            config.set_value(source, name)

        value = config.get_value(source)
        # if never set but called, then through exception
        if value is None:
            logging.error('Trying to inject dependency of unresolved name: {}'.format(property))
            raise DependencyInjectionError()

        setattr(obj_instance, property, value)
        return func(*args, **kwargs)
    return wrapper


def config(func=None, path: str = None, property: str = None):
    """
    Config decorator to import and inject the specified config.
    This config instance can be grouped in a defined name.
    The default config containing this new property is injected into an object instance.
    If none of the options is set, then only the default config if injected.
    :param func: original function instance
    :param path: path of the config file
    :param property: source name to categorize new config instance
    :return:
    """
    if not func:
        # workaround to enable empty decorator
        return functools.partial(config, path=path, property=property)

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        _handle_config(args, path, property)
        return func(*args, **kwargs)
    return wrapper


def context(func):
    """
    Context decorator to inject the default context instance.
    :param func: original function do inject instances
    :return:
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        context_ = Context.get_instance()
        config = Config.get_instance()
        registry = Registry.get_instance()
        search = ParameterSearch.get_instance()
        instance = args[0]

        setattr(instance, 'context', context_)
        setattr(instance, 'config', config)
        setattr(instance, 'registry', registry)
        setattr(instance, 'search', search)
        return func(*args, **kwargs)
    return wrapper


class InjectOption(Enum):
    Type = 0
    Instance = 1
    Config = 2
    Search = 3


def inject(source: str, property: str, option: InjectOption = InjectOption.Instance):
    """
    Injects a named variable of any given instance into the decorated object instance.
    :param source: Source instance available in the registry.
    :param property: Target name of the variable to inject.
    :param option: Option for the registry lookup.
    :return:
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            context = Context.get_instance()
            obj_instance = args[0]

            # if instances option selected lookup attribute or load if not already instantiated
            if option == InjectOption.Instance:
                registry = Registry.get_instance()
                instance, key = DotDict.resolve(registry.instances, source)
                value = getattr(instance, key, None)
                # prevent all side effects if allow_context_changes is False
                if context.allow_context_changes and value is None:
                    value = _search_and_load_type(source)
            # otherwise if types then return simple type
            elif option == InjectOption.Type:
                registry = Registry.get_instance()
                instance, key = DotDict.resolve(registry.types, source)
                value = getattr(instance, key, None)
            elif option == InjectOption.Config:
                config = Config.get_instance()
                instance, key = DotDict.resolve(config, source)
                value = getattr(instance, key, None)
            elif option == InjectOption.Search:
                search = ParameterSearch.get_instance()
                instance, key = DotDict.resolve(search, source)
                value = getattr(instance, key, None)
            else:
                raise NotImplementedError()

            # if value is still None then through error, since it was never registered or failed to load
            if value is None:
                logging.error('Trying to inject dependency of unresolved name: {}'.format(property))
                raise DependencyInjectionError()

            setattr(obj_instance, property, value)
            return func(*args, **kwargs)
        return wrapper
    return decorator


def register(type: str, property: str = None):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            context = Context.get_instance()
            # disable register to prevent all side effects on the registry
            if not context.allow_context_changes:
                return func(*args, **kwargs)

            obj_instance = args[0]

            registry = Registry.get_instance()

            instance, key = DotDict.resolve(registry.instances, type)
            setattr(instance, key, type)

            registry.register_type(key, type)
            module = Loader.import_path(type)
            value = module()
            registry.register_instance(key, value)

            if property is not None:
                setattr(obj_instance, property, value)

            return func(*args, **kwargs)
        return wrapper
    return decorator


def hook(method: str, replace_with: Callable, *args, **kwargs):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args_, **kwargs_):
            result = func(*args_, **kwargs_)
            instance = args_[0]
            m = getattr(instance, method)
            setattr(instance, method, lambda: replace_with(m, *args, **kwargs))
            return result
        return wrapper
    return decorator


def strategy(config: str, source: str = 'strategy', properties: list = None, ignore_none_values: bool = True):
    """
    Strategy decorator to inject a training strategy instance.
    :param source: defines the source name of the current training strategy
    :param config: property file for the current training strategy
    :param properties: names of the injection variable as specified in the config
    :param ignore_none_values: ignores None value references of injections
    :return:
    """
    if properties is None:
        properties = DEFAULT_PROPERTIES

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            _handle_config(args, config, source)
            _handle_registration(source)
            _handle_injections(args, properties, ignore_none_values=ignore_none_values)
            return func(*args, **kwargs)
        return wrapper
    return decorator


def reference(name: str):
    """
    Reference decorator to inject multiple values at once without registering new objects.
    :param name: Property to inject.
    :return:
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            _handle_injections(args, [name], ignore_none_values=False)
            return func(*args, **kwargs)
        return wrapper
    return decorator


def references(func=None, names: list = None, ignore_none_values: bool = True):
    """
    Reference decorator to inject multiple values at once without registering new objects.
    If no list is provided the default properties will be injected:
    'transform', 'dataset', 'data_builder', 'model', 'optimizer', 'collectible', 'criterion', 'metric', 'writer'
    :param func: original function instance
    :param names: List of properties to inject.
    :param ignore_none_values: ignores None value references of injections
    :return:
    """
    if names is None:
        names = DEFAULT_PROPERTIES
    if not func:
        # workaround to enable empty decorator
        return functools.partial(references, names=names, ignore_none_values=ignore_none_values)
    return _wrapper_delegate(func, names, ignore_none_values=ignore_none_values)


def transform(func):
    """
    Transform decorator to inject the default transform instance.
    :param func: original function instance
    :return:
    """
    properties = ['transform']
    if not func:
        # workaround to enable empty decorator
        return functools.partial(transform)
    return _wrapper_delegate(func, properties, ignore_none_values=False)


def dataset(func):
    """
    Dataset decorator to inject the default dataset instance.
    :param func: original function instance
    :return:
    """
    properties = ['dataset']
    if not func:
        # workaround to enable empty decorator
        return functools.partial(dataset)
    return _wrapper_delegate(func, properties, ignore_none_values=False)


def model(func):
    """
    Model decorator to inject the default model instance.
    :param func: original function instance
    :return:
    """
    properties = ['model']
    if not func:
        # workaround to enable empty decorator
        return functools.partial(model)
    return _wrapper_delegate(func, properties, ignore_none_values=False)


def metric(func):
    """
    Metric decorator to inject the default metric instance.
    :param func: original function instance
    :return:
    """
    properties = ['metric']
    if not func:
        # workaround to enable empty decorator
        return functools.partial(metric)
    return _wrapper_delegate(func, properties, ignore_none_values=False)


def search(group: str = None, params: List[Tuple[str, Parameter]] = None):
    """
    Decorator for searching hyper-parameters.
    :param group: collects the params to search only in an grouped logical search space
    :param params: list of tuples containing searchable parameters
    :return:
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            search = ParameterSearch.get_instance()
            context = Context.get_instance()
            instance = args[0]

            setattr(instance, 'search', search)
            # prevent all side effects on search
            if not context.allow_context_changes:
                return func(*args, **kwargs)

            if group is None and params is None:
                return func(*args, **kwargs)

            if group is None:
                property = search
            else:
                if group not in search.keys():
                    property = Box()
                    search[group] = property
                property = search[group]

            if params is not None:
                for param in params:
                    property[param[0]] = param[1]
            return func(*args, **kwargs)
        return wrapper
    return decorator



import torch
import numpy as np


class XdA(torch.nn.Module):
    """
    Context-dependent activations implementation introduced by Dinu et al. in the thesis
    Overcoming Catastrophic Forgetting with Context-Dependent Activations and Synaptic Stabilization
    https://www.dinu.at/wp-content/uploads/2019/11/Overcoming-Catastrophic-Forgetting-with-Context-Dependent-Activations-and-Synaptic-Stabilization.pdf
    track the layer activation statistics and use the summary statistics for the gating non-linearity.
    All activation functions that are based-on activation space statistics are denoted as context-dependent activations.
    This XdA class implementation currently supports two summary statistic variants, 'mean' and 'median' and the
    gating function is defined as followed:
    f(x) = | x >= phi  --> x (identity)
           | otherwise --> beta * x
    and
    phi <- (1 - alpha) * phi + alpha * x_bar
    whereas phi is the cumulative normalized summary statistic, x_bar the current normalized summary statistics of layer
    activations - which is here updated as an exponential moving mode - x are the input activations and
    beta and alpha are hyperparameters for tuning the updates and gating function.
    This gating function uses LayerNorm to re-normalize the input.
    """
    def __init__(self, alpha: float = 0.01,
                 beta: float = 0.0,
                 mode: str = 'mean',
                 norm_type: str = 'layernorm',
                 mask_type: str = 'mask',
                 mse_lr: float = 1e-2):
        """
        Context-dependent activations implementation using normalized activation space summary statistics
        for the non-linearity definition.
        The activation function initializes itself with the first passed data batch.
        ATTENTION: An XdA layer is NOT stateless, such as ReLU or other activation function. You need to instantiate one
        XdA-layer per output layer fot the gating to properly work.
        :param alpha: The summary statistic is updated via an exponential moving mode. The alpha trades of between how
        much the new current vs the new activation summary statistic contributes.
        :param beta: The beta parameter regulates the leakyness of the gating non-linearity.
        :param mode: The summary statistic
        :param norm_type: Normalize the x values according to the given normalization type before activation
        :param mse_lr: Learning rate for the differential parameters in the 'mse' mode.
        :param mask_type: Select the gating mode.
        """
        super(XdA, self).__init__()
        self.alpha = alpha
        self.beta = beta
        self.mode = mode
        self.phi = None
        self.norm = None
        self.norm_type = norm_type
        self.mask_type = mask_type
        self.mse_lr = mse_lr
        assert 0 <= self.alpha <= 1
        assert 0 <= self.beta <= 1
        assert 0 <= self.mse_lr
        assert self.mode in ['zero', 'mean', 'median', 'mse']
        assert self.norm_type in [None, 'vecnorm', 'batchnorm', 'layernorm', 'multinorm', 'mean', 'std']
        assert self.mask_type in ['mask', 'max']

    @staticmethod
    def _multi_norm(x, eps=1e-7):
        """
        Computes the batch and layer norm for the committed value x.
        """
        b_mean = torch.mean(x, dim=0, keepdim=True)
        b_std = torch.sqrt(torch.var(x, dim=0, keepdim=True) + eps)
        x = (x - b_mean) / b_std
        x_shape = x.shape
        if len(x_shape) == 2:
            x = (x - torch.mean(x, dim=-1, keepdim=True)) / torch.sqrt(torch.var(x, dim=-1, keepdim=True) + eps)
        elif len(x_shape) > 2:
            x = x.view(x_shape[0], -1)
            x = (x - torch.mean(x, dim=-1, keepdim=True)) / torch.sqrt(torch.var(x, dim=-1, keepdim=True) + eps)
            x = x.view(x_shape)
        else:
            raise NotImplementedError('Unsupported normalization shape: {}'.format(x_shape))
        return x

    @staticmethod
    def _layer_norm(x, eps=1e-7):
        """
        Computes the layer norm for the committed value x.
        """
        x_shape = x.shape
        if len(x_shape) == 2:
            x = (x - torch.mean(x, dim=-1, keepdim=True)) / torch.sqrt(torch.var(x, dim=-1, keepdim=True) + eps)
        elif len(x_shape) > 2:
            x = x.view(x_shape[0], -1)
            x = (x - torch.mean(x, dim=-1, keepdim=True)) / torch.sqrt(torch.var(x, dim=-1, keepdim=True) + eps)
            x = x.view(x_shape)
        else:
            raise NotImplementedError('Unsupported normalization shape: {}'.format(x_shape))
        return x

    @staticmethod
    def _batch_norm(x, eps=1e-7):
        """
        Computes the batch norm for the committed value x.
        """
        b_mean = torch.mean(x, dim=0, keepdim=True)
        b_std = torch.sqrt(torch.var(x, dim=0, keepdim=True) + eps)
        x = (x - b_mean) / b_std
        return x

    @staticmethod
    def _vec_norm(x, p=2, eps=1e-7):
        """
        Computes the normalized vector for the committed value x.
        """
        x_shape = x.shape
        if len(x_shape) == 2:
            x_norm = torch.pow(torch.sum(torch.pow(x, p), dim=-1, keepdim=True) + eps, 1/p)
            x = x / x_norm
        elif len(x_shape) > 2:
            x = x.view(x_shape[0], -1)
            x_norm = torch.pow(torch.sum(torch.pow(x, p), dim=-1, keepdim=True) + eps, 1/p)
            x = x / x_norm
            x = x.view(x_shape)
        else:
            raise NotImplementedError('Unsupported normalization shape: {}'.format(x_shape))
        return x

    @staticmethod
    def _mean(x):
        """
        Computes the mean normalization for a given x.
        """
        x_shape = x.shape
        if len(x_shape) == 2:
            x = x - torch.mean(x, dim=1, keepdim=True)
        elif len(x_shape) > 2:
            x = x.view(x_shape[0], -1)
            x = x - torch.mean(x, dim=1, keepdim=True)
            x = x.view(x_shape)
        else:
            raise NotImplementedError('Unsupported normalization shape: {}'.format(x_shape))
        return x

    @staticmethod
    def _std(x, eps=1e-7):
        """
        Computes the mean normalization for a given x.
        """
        x_shape = x.shape
        if len(x_shape) == 2:
            x_std = torch.sqrt(torch.var(x, dim=1, keepdim=True) + eps)
            x = x / x_std
        elif len(x_shape) > 2:
            x = x.view(x_shape[0], -1)
            x_std = torch.sqrt(torch.var(x, dim=1, keepdim=True) + eps)
            x = x / x_std
            x = x.view(x_shape)
        else:
            raise NotImplementedError('Unsupported normalization shape: {}'.format(x_shape))
        return x

    def _check_init(self, x):
        """
        Checks if the activation function is already initialized and if not, it creates the respective
        methods.
        :param x: the activation space inputs
        :return:
        """
        if self.phi is None:
            # init normalization
            if self.norm_type == 'vecnorm':
                self.norm = XdA._vec_norm
            elif self.norm_type == 'batchnorm':
                self.norm = XdA._layer_norm
            elif self.norm_type == 'layernorm':
                self.norm = XdA._layer_norm
            elif self.norm_type == 'multinorm':
                self.norm = XdA._multi_norm
            elif self.norm_type == 'mean':
                self.norm = XdA._mean
            elif self.norm_type == 'std':
                self.norm = XdA._std
            # init phi parameters
            if self.mode == 'zero':
                self.phi = torch.nn.Parameter(torch.zeros(x.shape[1:]), requires_grad=False).to(x.device)
            elif self.mode == 'mean':
                self.phi = torch.nn.Parameter(torch.zeros(x.shape[1:]), requires_grad=False).to(x.device)
                x_ = x.detach().mean(0).data
                self.phi.data = self.phi + x_
            elif self.mode == 'median':
                self.phi = torch.nn.Parameter(torch.zeros(x.shape[1:]), requires_grad=False).to(x.device)
                x_ = torch.median(x.detach(), dim=0)[0]
                self.phi.data = self.phi + x_
            elif self.mode == 'mse':
                self.phi = torch.from_numpy(np.random.randn(*x.shape[1:]).astype(np.float32)).to(x.device)
                self.phi.requires_grad = True
            self.init = False

    def _update_phi(self, x):
        """
        Updates the gating summary statistics according to the defined mode ('mean', 'median').
        :param x: the activation space inputs
        :return:
        """
        if self.training:
            if self.mode == 'mean':
                x_ = x.detach().mean(0)
                self.phi.data = ((1 - self.alpha) * self.phi + self.alpha * x_).data
            elif self.mode == 'median':
                x_ = torch.median(x.detach(), dim=0)[0]
                self.phi.data = ((1 - self.alpha) * self.phi + self.alpha * x_).data
            elif self.mode == 'mse':
                self.phi.data -= self.mse_lr * (self.phi - x.detach()).mean()

    def _mask(self, x):
        """
        Creates a binary mask according to the cumulative summary statistics.
        :param x: teh activation space inputs
        :return: binary mask with same dimensions as the inputs
        """
        return (x >= self.phi.expand_as(x)).float()

    def _gate(self, x):
        if self.mask_type == 'mask':
            # creates a binary mask to disable units
            mask = self._mask(x)
            # disables units
            h = x * mask
            # checks if beta is enabled and applies a leaky variant
            if self.beta != 0:
                h += (self.beta * x * (1.0 - mask))
        elif self.mask_type == 'max':
            if self.beta == 0:
                h = torch.max(x, self.phi.expand_as(x))
            else:
                h = torch.max(x, self.beta * self.phi.expand_as(x))
        else:
            raise NotImplementedError('Invalid option!')
        return h

    def forward(self, x):
        # check if layer is initialized
        self._check_init(x)
        # normalize the layer according to the layer norm
        if self.norm is not None:
            x = self.norm(x)
        # updates the gating summary statistic
        self._update_phi(x)
        h = self._gate(x)
        return h



import importlib
import traceback
import logging

from threading import RLock
from lighter.registry import Registry


class Loader(object):
    _instance = None
    # mutex for threadsafe instance loading
    _mutex = RLock()

    @staticmethod
    def import_modules(modules: dict):
        """
        Registers a dictionary of modules to the registry.
        :param modules: list of types
        :return:
        """
        for k, v in modules.items():
            if isinstance(v, str):
                module = Loader.import_path(v)
                setattr(Registry.get_instance().types, k, module)

    @staticmethod
    def import_path(name: str):
        """
        Imports a defined python type.
        :param name: path to a python module
        :return:
        """
        type_ = None
        try:
            components = name.split('.')
            module = importlib.import_module('.'.join(components[:-1]))
            type_ = getattr(module, components[-1])
        except Exception as e:
            tb = traceback.format_exc()
            logging.warning('Loader: Could not import: {} - {}'.format(name, e))
            logging.warning(tb)
        return type_



from lighter.context import Context
from lighter.decorator import register


class Runner:
    @register(type='experiments.mnist.Experiment', property='experiment')
    def __init__(self):
        pass

    def run(self):
        self.experiment.run()


def exec(config='configs/mnist.config.json'):
    Context.create(config)
    runner = Runner()
    runner.run()



from torchvision.transforms import Compose


class BaseTransform(Compose):
    """Base class for transforming data.
    """
    def __init__(self, transforms):
        super(BaseTransform, self).__init__(transforms)



import torch
import torch.nn as nn


class DynLayerNorm(nn.Module):
    """
    Computes the layer norm for the committed value x. The dimensions are automatically established.
    """
    def __init__(self):
        super(DynLayerNorm, self).__init__()

    @staticmethod
    def _layer_norm(x, eps=1e-7):
        x_shape = x.shape
        if len(x_shape) == 2:
            x = (x - torch.mean(x, dim=-1, keepdim=True)) / torch.sqrt(torch.var(x, dim=-1, keepdim=True) + eps)
        elif len(x_shape) > 2:
            x = x.view(x_shape[0], -1)
            x = (x - torch.mean(x, dim=-1, keepdim=True)) / torch.sqrt(torch.var(x, dim=-1, keepdim=True) + eps)
            x = x.view(x_shape)
        else:
            raise NotImplementedError('Unsupported normalization shape: {}'.format(x_shape))
        return x

    def forward(self, x):
        return self._layer_norm(x)


class DynBatchNorm(nn.Module):
    """
    Computes the batch and layer norm for the committed value x. The dimensions are automatically established.
    """
    def __init__(self):
        super(DynBatchNorm, self).__init__()

    @staticmethod
    def _multi_norm(x, eps=1e-7):
        b_mean = torch.mean(x, dim=0, keepdim=True)
        b_std = torch.sqrt(torch.var(x, dim=0, keepdim=True) + eps)
        x = (x - b_mean) / b_std
        return x

    def forward(self, x):
        return self._multi_norm(x)




from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    readme = fh.read()

requirements = {"install": ["ffmpeg", "natsort", "numpy", "tqdm", "torch", "torchvision",
                            "tensorboard", "coloredlogs", "petname", "pandas", "ray", "setproctitle",
                            "multiprocess", "python-box"]}

install_requires = requirements["install"]

setup(
    # Metadata
    name="torch-lighter",
    author="Marius-Constantin Dinu",
    version="0.2.20",
    author_email="dinu.marius-constantin@hotmail.com",
    url="https://github.com/Xpitfire/lighter",
    scripts=['bin/lighter-init'],
    description="Lightweight extension for torch to speed-up prototyping.",
    long_description=readme,
    long_description_content_type="text/markdown",
    # Package info
    packages=find_packages(),
    zip_safe=True,
    install_requires=install_requires,
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent"
    ]
)



import os
import json
import torch
from torch.multiprocessing import Process
from lighter.context import Context
from lighter.loader import Loader


class ContextBuilder(object):
    """
    Is used to create a proper context object for running an experiment with the current
    """
    def __init__(self,
                 process_id: str,
                 schedule_file: str,
                 experiment: str,
                 device: str = 'cpu'):
        self.process_id = process_id
        self.schedule_file = schedule_file
        self.device = device
        if 'cuda:' in self.device:
            torch.cuda.set_device(int(self.device.split(':')[-1]))
        with open(schedule_file) as json_file:
            schedule = json.load(json_file)
        self.files = schedule[self.process_id]
        self.experiment = Loader.import_path(experiment)

    def __iter__(self):
        self.idx = 0
        return self

    def __next__(self):
        """
        Iterates over the schedules list of config files and prepares the context for execution.
        :return: a prepared experiment
        """
        if self.idx < len(self.files):
            # create new context with default config
            context = Context.create(config_file=self.files[self.idx],
                                     parse_args_override=True,
                                     auto_instantiate_types=False,
                                     allow_context_changes=False)

            # assign a new default device
            context.config.device.default = self.device
            # assign the process_id
            context.config.set_value('process_id', self.process_id)

            # create types
            context.instantiate_types(context.registry.types)

            # create a new experiment
            experiment = self.experiment()

            self.idx += 1
            return experiment
        else:
            raise StopIteration


class Executor:
    """
    Defines the main worker for executing a schedule flow.
    """
    @staticmethod
    def worker(process_id: str, schedule_file: str, experiment: str, device):
        scheduler = ContextBuilder(process_id=process_id,
                                   schedule_file=schedule_file,
                                   experiment=experiment,
                                   device=device)
        for i, exp in enumerate(scheduler):
            exp()


class Scheduler:
    """
    Schedules a list of parameters within a designated path for execution on multiple devices.
    """
    def __init__(self,
                 path: str,
                 experiment: str,
                 device_name: str = 'cuda',
                 num_workers: int = 1):
        self.path = path
        self.experiment = experiment
        self.device_name = device_name
        self.num_workers = num_workers
        self.schedule_file = None
        self.processes = []

    def build_schedule(self):
        """
        Builds a schedule file from the defined path with configs.
        :return:
        """
        # create empty schedule
        schedule = {'{}:{}'.format(self.device_name, i % self.num_workers): [] for i in range(self.num_workers)}
        # assign configs to schedule
        for i, file in enumerate(os.listdir(self.path)):
            schedule['{}:{}'.format(self.device_name, i % self.num_workers)].append(
                os.path.join(self.path, file))
        # save schedule
        self.schedule_file = os.path.join(self.path, 'schedule.json')
        dict_str = json.dumps(schedule, indent=2)
        with open(self.schedule_file, 'w') as file:
            file.write(dict_str)

    def create_processes(self):
        """
        Creates a list of processes with execution workers.
        :return:
        """
        for i in range(self.num_workers):
            process_id = '{}:{}'.format(self.device_name, i)
            device = process_id
            if 'cpu' in device:
                device = 'cpu'
            process = Process(name=process_id,
                              target=Executor.worker,
                              args=(process_id, self.schedule_file, self.experiment, device))
            self.processes.append(process)

    def execute_processes(self):
        """
        Executes a schedule on the initiated process workers.
        :return:
        """
        for t in self.processes:
            t.start()
        [p.join() for p in self.processes]

    def clean_processes(self):
        """
        Deletes the list of processes.
        :return:
        """
        self.processes = []

    def run(self):
        """
        Main execution loop for starting a scheduled run.
        :return:
        """
        self.build_schedule()
        self.create_processes()
        self.execute_processes()
        self.clean_processes()



from typing import List
from box import Box
from lighter.config import Config
from threading import RLock
from lighter.misc import generate_long_id
from lighter.parameter import Parameter


class ParameterSearch(Box):
    """
    Parameter search collection updated by the @search decorator.
    """
    _mutex = RLock()
    _instance = None

    def __init__(self, **kwargs):
        super(ParameterSearch, self).__init__(**kwargs)

    @staticmethod
    def compile(params: List["Parameter"], config: Config = None) -> List[Config]:
        configs = []
        if len(params) <= 0:
            # fix to get unique name
            config['experiment_id'] = generate_long_id()
            return [config]
        param = params[0]
        if config is not None:
            param.update_config(config)
        for p in param:
            configs = configs + ParameterSearch.compile(params[1:], p)
        return configs

    @staticmethod
    def get_instance() -> "ParameterSearch":
        ParameterSearch._mutex.acquire()
        try:
            return ParameterSearch._instance
        finally:
            ParameterSearch._mutex.release()

    @staticmethod
    def create_instance() -> "ParameterSearch":
        ParameterSearch._mutex.acquire()
        try:
            ParameterSearch._instance = ParameterSearch()
            return ParameterSearch._instance
        finally:
            ParameterSearch._mutex.release()



from lighter.decorator import context


class BaseOptimizer(object):
    """
    Base class proxy for the PyTorch optimizers with injected 'model' and 'context' instances.
    """
    @context
    def __init__(self):
        self.optimizer = None

    def state_dict(self):
        return self.optimizer.state_dict()

    def load_state_dict(self, state_dict):
        self.optimizer.state_dict(state_dict)

    def zero_grad(self):
        self.optimizer.zero_grad()

    def step(self, closure=None):
        self.optimizer.step(closure)

    def add_param_group(self, param_group):
        self.optimizer.add_param_group(param_group)
